from SurvSet.data import SurvLoader
import pandas as pd

from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import StandardScaler, OneHotEncoder
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.model_selection import StratifiedShuffleSplit
from sklearn.utils import shuffle

from tte_utils.dataset import MyDataset


def load_dataset(dataset_name: str):
    """
    :param dataset_name:
    :return: Dataframe, reference to the dataset
    """
    loader = SurvLoader()
    df, ref = loader.load_dataset(ds_name=dataset_name).values()
    return df, ref


def preprocess_and_split(df: pd.DataFrame, perform_z_score: bool = False):
    """
    Preprocess the data
    :param df: dataframe
    :param perform_z_score: boolean
    :return: preprocessed dataframe
    """

    # drop pid column
    df = df.drop(['pid'], axis=1)

    df = shuffle(df, random_state=42)
    df.reset_index(drop=True, inplace=True)

    split = StratifiedShuffleSplit(n_splits=1, test_size=0.2, random_state=42)

    for train_index, test_index in split.split(df, df['event']):
        df_train, df_test = df.iloc[train_index], df.iloc[test_index]

    # Define transformers for categorical and numerical features
    categorical_features = df.columns[df.columns.str.startswith('fac')]
    numerical_features = df.columns[df.columns.str.startswith('num')]
    unchanged_features = ['event', 'time']

    # Pipeline for categorical features (Imputation + OneHotEncoding)
    categorical_pipeline = Pipeline([
        ('imputer', SimpleImputer(strategy='most_frequent', missing_values=None)),
        ('onehot', OneHotEncoder(handle_unknown='ignore', sparse_output=False))
    ])

    # Pipeline for numerical features (Imputation + Z-score scaling)
    numerical_pipeline = Pipeline([
        ('imputer', SimpleImputer(strategy='mean'))
    ])

    # Combine pipelines using ColumnTransformer
    preprocessor = ColumnTransformer([
        ('cat', categorical_pipeline, categorical_features),
        ('num', numerical_pipeline, numerical_features),
        ('unchanged', 'passthrough', unchanged_features)
    ])

    # Full pipeline with missing value handling and preprocessing
    full_pipeline = Pipeline([
        ('missing_handler', MissingValueHandler(df=df_train)),
        ('preprocessor', preprocessor),
    ])

    # Fit-transform the train data
    transformed_data_train = full_pipeline.fit_transform(df_train)

    # Get the feature names after transformation
    if len(categorical_features) > 0:
        categorical_feature_names = full_pipeline.named_steps['preprocessor'].named_transformers_['cat'][
            'onehot'].get_feature_names_out(categorical_features)
        categorical_feature_names = list(categorical_feature_names)
    else:
        categorical_feature_names = []
    numerical_feature_names = numerical_features
    all_feature_names = list(categorical_feature_names) + list(numerical_feature_names) + unchanged_features

    # Convert to DataFrame
    transformed_df_train = pd.DataFrame(transformed_data_train, columns=all_feature_names)

    # print(transformed_df_train)

    # Transform the test data
    transformed_data_test = full_pipeline.transform(df_test)
    transformed_df_test = pd.DataFrame(transformed_data_test, columns=all_feature_names)
    # print(transformed_df_test)
    if perform_z_score:
        features_to_normalize = transformed_df_train.columns[transformed_df_train.columns.str.startswith('num') + transformed_df_train.columns.str.startswith('time')]
        transformed_df_train, numerical_pipeline = perform_z_score_on_numerical_features(transformed_df_train, features_to_normalize)
        transformed_df_test[features_to_normalize] = numerical_pipeline.transform(transformed_df_test[features_to_normalize])

    X_train = transformed_df_train.drop(['event', 'time'], axis=1)
    y_train = transformed_df_train[['event', 'time']]

    X_test = transformed_df_test.drop(['event', 'time'], axis=1)
    y_test = transformed_df_test[['event', 'time']]

    train_dataset = MyDataset(X_train, y_train)
    test_dataset = MyDataset(X_test, y_test)

    return train_dataset, test_dataset


def perform_z_score_on_numerical_features(df: pd.DataFrame, features_to_normalize):
    """
    Perform Z-score scaling on numerical features
    :param df: dataframe
    :return: scaled dataframe
    """
    numerical_pipeline = Pipeline([
        ('scaler', StandardScaler())
    ])

    df[features_to_normalize] = numerical_pipeline.fit_transform(df[features_to_normalize])

    return df, numerical_pipeline


class MissingValueHandler(BaseEstimator, TransformerMixin):
    def __init__(self, df, threshold=0.5):
        self.threshold = threshold
        self.df = df

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        amount_of_rows_with_missing_values = self.df.isnull().any(axis=1).sum()
        missing_ratio = amount_of_rows_with_missing_values / self.df.shape[0]
        # print(f'Missing ratio: {missing_ratio}')
        # print(f'Threshold: {self.threshold}')
        if missing_ratio > self.threshold:
            return X.dropna()
        else:
            return X



from lifelines.utils import concordance_index
import torch

from tte_utils.dataset import MyDataset


def calculate_ci(model, val_dataset) -> float:
    """
    Calculate the concordance index
    :param model: Subclass of AbstractNetwork
    :param val_dataset: Dataset
    :return:
    """
    model.network.eval()
    y_val = val_dataset.y.to(model.device)

    with torch.no_grad():
        event_times = y_val[:, 1].flatten().detach().cpu().numpy()

        output = model.predict(val_dataset.X)
        if isinstance(output, tuple):
            t_pred, p_pred = output
        else:
            t_pred = output

        preds = list(t_pred.flatten().detach().cpu().numpy())

        # get the uncensored and censored samples as a numpy array
        uncensored_idx = MyDataset.is_uncensored(val_dataset.y).flatten().detach().cpu().numpy()
        # get the event types
        event_indicators = list(uncensored_idx)

    return concordance_index(event_times, preds, event_indicators)



import torch
from torch.utils.data import Dataset
import numpy as np


class MyDataset(Dataset):

    def __init__(self, X, y, is_train=True, mu=None, sigma=None):
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        X = X.to_numpy()  # size: (n_samples, n_features)
        y = y.to_numpy()  # size: (n_samples, 2): first column is the event time, second column is the event type

        X = torch.from_numpy(X.astype(np.float32)).to(self.device)
        y = torch.from_numpy(y.astype(np.float32)).to(self.device)

        self.X = X
        self.y = y

        # scale the features, if is_train=True, calculate mu and sigma, else use the given mu and sigma
        # if is_train:
        #     self.mu = X.mean(0, keepdim=True)
        #     self.sigma = X.std(0, unbiased=False, keepdim=True)
        #     self.X = (X - self.mu) / self.sigma
        # else:
        #     self.mu = mu
        #     self.sigma = sigma
        #     self.X = (X - self.mu) / self.sigma
        self.n_samples = X.shape[0]

    def __getitem__(self, index):
        return self.X[index], self.y[index]

    def __len__(self):
        return self.n_samples

    @staticmethod
    def is_censored(y):
        """
        :return: all the samples
        """
        if len(y.shape) == 1:
            val = y[0] == 0.0
            return torch.tensor([val])
        return y[:, 0] == 0.0

    @staticmethod
    def is_uncensored(y):
        """
        :return: all the samples
        """
        if len(y.shape) == 1:
            val = y[0] != 0.0
            return torch.tensor([val])
        return y[:, 0] != 0.0

    @staticmethod
    def get_time(y, device=None):
        # either the event time or the censored time, depends on the event type
        if len(y.shape) == 1:
            val = y[0]
            if device is None:
                device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            return torch.tensor([val]).to(device)
        return y[:, 1]



from sklearn.utils import shuffle
from sklearn.model_selection import StratifiedShuffleSplit
import pandas as pd
from tte_utils.dataset import MyDataset


def train_val_test_split(df: pd.DataFrame):
    """
    Split the data into train, validation and test sets
    :param df: dataframe
    :return: train, validation and test dataset
    """
    df = shuffle(df, random_state=42)
    df.reset_index(drop=True, inplace=True)
    X = df.drop(['pid', 'event', 'time'], axis=1)
    y = df[['event', 'time']]

    split = StratifiedShuffleSplit(n_splits=1, test_size=0.2, random_state=42)

    # reset index so that stratified shuffle split will work
    df.reset_index(drop=True, inplace=True)

    for dev_index, test_index in split.split(df, df['event']):
        X_dev, X_test = X.iloc[dev_index], X.iloc[test_index]
        y_dev, y_test = y.iloc[dev_index], y.iloc[test_index]

    split = StratifiedShuffleSplit(n_splits=1, test_size=0.2, random_state=42)

    # concat X_dev and y_dev
    df_dev = pd.concat([X_dev, y_dev], axis=1)

    for train_index, val_index in split.split(df_dev, df_dev['event']):
        X_train, X_val = X_dev.iloc[train_index], X_dev.iloc[val_index]
        y_train, y_val = y_dev.iloc[train_index], y_dev.iloc[val_index]

    train_dataset = MyDataset(X_train, y_train)
    val_dataset = MyDataset(X_val, y_val)
    test_dataset = MyDataset(X_test, y_test)

    return train_dataset, val_dataset, test_dataset


def train_test_split(df: pd.DataFrame):
    """
    Split the data into train, validation and test sets
    :param df: dataframe
    :return: train and test dataset
    """
    df = shuffle(df, random_state=42)
    df.reset_index(drop=True, inplace=True)
    X = df.drop(['pid', 'event', 'time'], axis=1)
    y = df[['event', 'time']]

    split = StratifiedShuffleSplit(n_splits=1, test_size=0.2, random_state=42)

    # reset index so that stratified shuffle split will work
    df.reset_index(drop=True, inplace=True)

    for train_index, test_index in split.split(df, df['event']):
        X_train, X_test = X.iloc[train_index], X.iloc[test_index]
        y_train, y_test = y.iloc[train_index], y.iloc[test_index]

    train_dataset = MyDataset(X_train, y_train)
    test_dataset = MyDataset(X_test, y_test)

    return train_dataset, test_dataset


from tte_models.ratio import RATIO
from tte_models.my_model import MyModel

RATIO_NAME = "RATIO"
MyModel_NAME = "MyModel"


def get_model(model_name: str, input_size: int):
    # TODO: allow for kwargs and checkpoint_dir
    model = None
    if model_name == RATIO_NAME:
        model = RATIO(input_size=input_size)
    elif model_name == MyModel_NAME:
        model = MyModel(input_size=input_size)

    return model



import unittest
from tte_models.ModelFactory import get_model


class TestGetModel(unittest.TestCase):

    def test_get_model(self):
        self.assertEquals(get_model("Some_Model", 10), None)


unittest.main()



import numpy as np
import torch.nn as nn
import torch
import torch.nn.functional as F
import pandas as pd
from torch.utils.data import DataLoader
from typing import Union

from tte_utils.dataset import MyDataset


class RATIO:
    def __init__(self,
                 input_size=1,
                 learning_rate=1e-3):
        self.input_size = input_size
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.network = RATIO_Network(input_size=input_size).to(self.device)
        self.optimizer = torch.optim.Adam(self.network.parameters(), lr=learning_rate)

    def fit(self, x: pd.DataFrame = None, y: pd.DataFrame = None,
            dataset: MyDataset = None,
            n_epochs=1000, batch_size=32, show_progress=True):
        """
        Train the model
        :param x: Dataframe
        :param y: Dataframe
        :param dataset: MyDataset: In case x and y are not provided
        :param n_epochs:
        :param batch_size:
        :param show_progress:
        :return:
        """

        self.network.train()

        # TODO: load coefficients
        if x is not None and y is not None:
            x = torch.from_numpy(x.to_numpy())
            y = torch.from_numpy(y.to_numpy())
            dataset = MyDataset(x, y)

        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=True)

        first_loss_history = []
        second_loss_history = []

        for epoch in range(n_epochs):
            # For the first loss
            first_loss_epoch = 0.0
            epoch_first_amount = 0
            # For the second loss
            second_loss_epoch = 0.0
            epoch_second_amount = 0

            for batch_x, batch_y in dataloader:
                batch_x = batch_x.to(self.device)
                batch_y = batch_y.to(self.device)

                self.optimizer.zero_grad()

                output = self.network(batch_x)

                # combined_loss is for back prop, the rest are for logging
                combined_loss, first_loss, batch_first_amount, second_loss, batch_second_amount = RATIO.ratio_loss(
                    y_pred=output, y_true=batch_y)

                combined_loss.backward()
                self.optimizer.step()

                with torch.no_grad():
                    first_loss_epoch += first_loss.item() * batch_first_amount
                    epoch_first_amount += batch_first_amount
                    second_loss_epoch += second_loss.item() * batch_second_amount
                    epoch_second_amount += batch_second_amount

            first_loss_history.append(first_loss_epoch / epoch_first_amount)
            second_loss_history.append(second_loss_epoch / epoch_second_amount)

            if show_progress:
                if epoch % 100 == 0:
                    print(f"Epoch {epoch}, First Loss: {first_loss_epoch / epoch_first_amount}, "
                          f"Second Loss: {second_loss_epoch / epoch_second_amount}")

        return first_loss_history, second_loss_history

    @staticmethod
    def ratio_loss(y_pred, y_true, coefficients=None, device=None):
        if device is None:
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        loss_first, loss_second = 0.0, 0.0

        if coefficients is None:
            coefficients = [1.0, 0.5]  # default parameters

        uncensored_idx = y_true[:, 0] == 1
        censored_idx = y_true[:, 0] == 0
        y_cen, y_hat_cen = y_true[censored_idx], y_pred[censored_idx]

        combined_loss = 0.
        # First loss component
        batch_first_amount = uncensored_idx.sum().item()
        if any(uncensored_idx):
            loss_first = F.mse_loss(y_true[uncensored_idx, 1].view(-1, 1), y_pred[uncensored_idx].view(-1, 1),
                                    reduction='mean') * coefficients[0]
            combined_loss += loss_first

        if len(y_cen) == 0:
            if isinstance(combined_loss, float):
                combined_loss = torch.tensor(0., torch.float32, requires_grad=True).to(device)
            return combined_loss

        # Second loss component
        to_change = y_hat_cen.flatten() < y_cen[:, 1].flatten()
        batch_second_amount = to_change.sum().item()
        if any(to_change):
            loss_second = F.mse_loss(y_cen[to_change, 1].view(-1, 1), y_hat_cen[to_change].view(-1, 1)) * coefficients[
                1]
            combined_loss += loss_second

        loss_first = loss_first.item() if not isinstance(loss_first, float) else 0
        loss_second = loss_second.item() if not isinstance(loss_second, float) else 0

        # Last part from ratio
        not_changed = len(y_cen) - sum(to_change)
        if (len(y_pred) - not_changed) != 0:
            combined_loss = (combined_loss * len(y_pred)) / (len(y_pred) - not_changed)

            # Fix loss components accordingly
            loss_first = loss_first * len(y_pred) / (len(y_pred) - not_changed)
            loss_second = loss_second * len(y_pred) / (len(y_pred) - not_changed)

        return combined_loss, loss_first, batch_first_amount, loss_second, batch_second_amount

    def predict(self, x: Union[pd.DataFrame, np.ndarray]):
        self.network.eval()
        with torch.no_grad():
            if isinstance(x, pd.DataFrame):
                x = torch.from_numpy(x.to_numpy()).float()
            else:
                x = torch.from_numpy(x).float()
            x = x.to(self.device)
            return self.network(x).cpu().detach().numpy()

    def predict_proba(self, x: Union[pd.DataFrame, np.ndarray]):
        raise NotImplementedError("Model does not predict probabilities")


class RATIO_Network(nn.Module):
    def __init__(self, input_size):
        super(RATIO_Network, self).__init__()

        self.input_size = input_size

        self.fc1 = nn.Linear(self.input_size, 1000)
        self.bn1 = nn.BatchNorm1d(1000)
        self.drop1 = nn.Dropout(0.2)
        self.relu1 = nn.ReLU()

        self.fc2 = nn.Linear(1000, 1000)
        self.bn2 = nn.BatchNorm1d(1000)
        self.drop2 = nn.Dropout(0.2)
        self.relu2 = nn.ReLU()

        self.fc3 = nn.Linear(1000, 1)

    def forward(self, x):
        out = self.fc1(x)
        # out = self.bn1(out)
        out = self.relu1(out)
        # out = self.drop1(out)

        out = self.fc2(out)
        # out = self.bn2(out)
        out = self.relu2(out)
        # out = self.drop2(out)

        out = self.fc3(out)

        t = out
        t = F.relu(t)

        return t



import numpy as np
import torch.nn as nn
import torch
import torch.nn.functional as F
import pandas as pd
from torch.utils.data import DataLoader
from typing import Union

from tte_utils.dataset import MyDataset


class MyModel:
    def __init__(self,
                 input_size=1,
                 learning_rate=1e-3):
        self.input_size = input_size
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.network = TwoNetworks(input_size=input_size).to(self.device)
        self.optimizer = torch.optim.Adam(self.network.parameters(), lr=learning_rate)

    def fit(self, x: pd.DataFrame = None, y: pd.DataFrame = None,
            dataset: MyDataset = None,
            n_epochs=1000, batch_size=32, show_progress=True):
        """
        Train the model
        :param x: Dataframe
        :param y: Dataframe
        :param dataset: MyDataset: In case x and y are not provided
        :param n_epochs:
        :param batch_size:
        :param show_progress:
        :return:
        """

        self.network.train()

        # TODO: load coefficients

        if x is not None and y is not None:
            x = torch.from_numpy(x.to_numpy())
            y = torch.from_numpy(y.to_numpy())
            dataset = MyDataset(x, y)

        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=True)

        first_loss_history = []
        second_loss_history = []
        third_loss_history = []

        uncensored_mean = np.mean(y[y[:, 0] == 1][:, 1]).to(self.device)
        # calculate the mean y value of the censored samples
        censored_mean = np.mean(y[y[:, 0] == 0][:, 1]).to(self.device)

        for epoch in range(n_epochs):
            # For the first loss
            first_loss_epoch = 0.0
            epoch_first_amount = 0
            # For the second loss
            second_loss_epoch = 0.0
            epoch_second_amount = 0
            # For the third loss
            third_loss_epoch = 0.0
            epoch_third_amount = 0

            for batch_x, batch_y in dataloader:
                batch_x = batch_x.to(self.device)
                batch_y = batch_y.to(self.device)

                self.optimizer.zero_grad()

                t_pred, p_pred = self.network(batch_x)

                # combined_loss is for back prop, the rest are for logging
                (combined_loss, first_loss, batch_first_amount, second_loss, batch_second_amount,
                 third_loss, batch_third_amount) = MyModel.my_loss(
                    t_pred, p_pred, y_true=batch_y,
                    uncensored_mean=uncensored_mean,
                    censored_mean=censored_mean)

                combined_loss.backward()
                self.optimizer.step()

                with torch.no_grad():
                    first_loss_epoch += first_loss.item() * batch_first_amount
                    epoch_first_amount += batch_first_amount

                    second_loss_epoch += second_loss.item() * batch_second_amount
                    epoch_second_amount += batch_second_amount

                    third_loss_epoch += third_loss.item() * batch_third_amount
                    epoch_third_amount += batch_third_amount

            first_loss_history.append(first_loss_epoch / epoch_first_amount)
            second_loss_history.append(second_loss_epoch / epoch_second_amount)
            third_loss_history.append(third_loss_epoch / epoch_third_amount)

            if show_progress:
                if epoch % 100 == 0:
                    print(f"Epoch {epoch}, First Loss: {first_loss_epoch / epoch_first_amount}, "
                          f"Second Loss: {second_loss_epoch / epoch_second_amount}")

        return first_loss_history, second_loss_history, third_loss_history

    @staticmethod
    def my_loss(t_pred, p_pred, y_true, uncensored_mean, censored_mean, coefficients=None, device=None):
        if device is None:
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        loss_first, loss_second, loss_third = 0.0, 0.0, 0.0

        if coefficients is None:
            coefficients = [100.0, 31.2, 0.5]  # default parameters

        uncensored_idx = y_true[:, 0] == 1
        censored_idx = y_true[:, 0] == 0

        combined_loss = 0.
        # if it remains zero, then loss is a float and will get converted to a tensor
        is_zero = True

        # First loss component: MSE loss for uncensored samples
        batch_first_amount = uncensored_idx.sum().item()
        if any(uncensored_idx):
            is_zero = False
            loss_first = F.mse_loss(y_true[uncensored_idx, 1].view(-1, 1), t_pred[uncensored_idx].view(-1, 1),
                                    reduction='mean') * coefficients[0]
            first_normalization = torch.sum(
                torch.square(MyDataset.get_time(y_true)[uncensored_idx].view(-1, 1) - uncensored_mean))
            loss_first /= first_normalization
            combined_loss += loss_first

        # Second loss component: negative log likelihood for uncensored samples
        batch_second_amount = uncensored_idx.sum().item()
        if any(uncensored_idx):
            loss_second = - torch.mean(torch.log(p_pred[uncensored_idx].view(-1, 1))) * coefficients[1]
        else:
            loss_second = torch.tensor(data=0., dtype=torch.float32, requires_grad=True)

        combined_loss += loss_second

        # calculate sum over the censored where we estimated lower than the censoring time: (t_pred - t_censored)^2 * p_pred
        # get the indices of the censored samples where the predicted time is lower than the censoring time
        to_change_idx = t_pred.flatten() < MyDataset.get_time(y_true).flatten()
        to_change_idx = to_change_idx & censored_idx

        batch_third_amount = to_change_idx.sum().item()
        if any(to_change_idx):
            is_zero = False

            first_term = torch.square(
                t_pred[to_change_idx].view(-1, 1) - MyDataset.get_time(y_true)[to_change_idx].view(-1, 1))
            second_term = p_pred[to_change_idx].view(-1, 1)

            loss_third = torch.mean(torch.mul(first_term, second_term)) * coefficients[2]
            third_normalization = torch.sum(
                torch.square(MyDataset.get_time(y_true)[censored_idx].view(-1, 1) - censored_mean))
            loss_third /= third_normalization

        if is_zero:
            combined_loss = torch.tensor(data=0., dtype=torch.float32, requires_grad=True).to(device)
            return combined_loss

        return combined_loss, loss_first, batch_first_amount, loss_second, batch_second_amount, loss_third, batch_third_amount

    def predict(self, x: Union[pd.DataFrame, np.ndarray]):
        self.network.eval()
        with torch.no_grad():
            if isinstance(x, pd.DataFrame):
                x = torch.from_numpy(x.to_numpy()).float()
            else:
                x = torch.from_numpy(x).float()
            x = x.to(self.device)
            return self.network(x).cpu().detach().numpy()

    def predict_proba(self, x: Union[pd.DataFrame, np.ndarray]):
        raise NotImplementedError("predict_proba is not implemented")


class TwoNetworks(nn.Module):
    def __init__(self, input_size):
        super(TwoNetworks, self).__init__()

        self.input_size = input_size

        self.t_fc1 = nn.Linear(self.input_size, 1000)
        # self.t_bn1 = nn.BatchNorm1d(1000)
        # self.t_drop1 = nn.Dropout(0.2)
        self.t_relu1 = nn.ReLU()

        self.t_fc2 = nn.Linear(1000, 1000)
        # self.t_bn2 = nn.BatchNorm1d(1000)
        # self.t_drop2 = nn.Dropout(0.2)
        self.t_relu2 = nn.ReLU()

        self.t_fc3 = nn.Linear(1000, 1)

        self.p_fc1 = nn.Linear(self.input_size, 1000)
        # self.p_bn1 = nn.BatchNorm1d(1000)
        # self.p_drop1 = nn.Dropout(0.2)
        self.p_relu1 = nn.ReLU()

        self.p_fc2 = nn.Linear(1000, 1000)
        # self.p_bn2 = nn.BatchNorm1d(1000)
        # self.p_drop2 = nn.Dropout(0.2)
        self.p_relu2 = nn.ReLU()

        self.p_fc3 = nn.Linear(1000, 1)

    def forward(self, x):
        t = self.t_fc1(x)
        t = self.t_relu1(t)
        t = self.t_fc2(t)
        t = self.t_relu2(t)
        t = self.t_fc3(t)
        t = F.relu(t)

        p = self.p_fc1(x)
        p = self.p_relu1(p)
        p = self.p_fc2(p)
        p = self.p_relu2(p)
        p = self.p_fc3(p)
        p = torch.sigmoid(p)

        return t, p

    def _common_step(self, x):
        out = self.fc1(x)
        # out = self.bn1(out)
        out = self.relu1(out)
        # out = self.drop1(out)

        out = self.fc2(out)
        # out = self.bn2(out)
        out = self.relu2(out)
        # out = self.drop2(out)
        out = self.fc3(out)
        return out
